/*++
/* NAME
/*	smtp_proto 3
/* SUMMARY
/*	client SMTP protocol
/* SYNOPSIS
/*	#include "smtp.h"
/*
/*	int	smtp_helo(state)
/*	SMTP_STATE *state;
/*
/*	int	smtp_xfer(state)
/*	SMTP_STATE *state;
/*
/*	void	smtp_quit(state)
/*	SMTP_STATE *state;
/* DESCRIPTION
/*	This module implements the client side of the SMTP protocol.
/*
/*	smtp_helo() performs the initial handshake with an SMTP
/*	server. It returns 0 in case of success, -1 in case of error.
/*
/*	smtp_xfer() sends message envelope information followed by the
/*	message data. The number of recipients per transaction is
/*	limited by the global variable \fIvar_smtp_rcpt_limit\fR.
/*	Recipients are marked as "done" in the mail queue file when
/*	bounced or delivered.
/*	The result value is 0 in case of success, -1 in case of error.
/*
/*	smtp_quit() performs the final protocol handshake. This
/*	function always claims success.
/* DIAGNOSTICS
/*	Warnings: corrupt message file. A corrupt message is marked
/*	as "corrupt" by changing its mode.
/* BUGS
/*	Some SMTP servers will abort when the number of recipients
/*	for one message exceeds their capacity. This behavior violates
/*	the SMTP protocol.
/*	The only way around this is to limit the number of recipients
/*	per transaction to an artificially-low value.
/* SEE ALSO
/*	smtp(3h) internal data structures
/*	smtp_chat(3) query/reply SMTP support
/*	smtp_trouble(3) error handlers
/* LICENSE
/* .ad
/* .fi
/*	The Secure Mailer license must be distributed with this software.
/* AUTHOR(S)
/*	Wietse Venema
/*	IBM T.J. Watson Research
/*	P.O. Box 704
/*	Yorktown Heights, NY 10598, USA
/*--*/

/* System library. */

#include <sys_defs.h>
#include <sys/stat.h>
#include <sys/socket.h>			/* shutdown(2) */
#include <setjmp.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>			/* 44BSD stdarg.h uses abort() */
#include <stdarg.h>

#ifdef STRCASECMP_IN_STRINGS_H
#include <strings.h>
#endif

/* Utility library. */

#include <msg.h>
#include <vstring.h>
#include <vstream.h>
#include <vstring_vstream.h>
#include <stringops.h>
#include <mymalloc.h>

/* Global library. */

#include <mail_params.h>
#include <smtp_stream.h>
#include <mail_queue.h>
#include <recipient_list.h>
#include <deliver_request.h>
#include <deliver_completed.h>
#include <defer.h>
#include <bounce.h>
#include <sent.h>
#include <record.h>
#include <rec_type.h>
#include <off_cvt.h>
#include <mark_corrupt.h>

/* Application-specific. */

#include "smtp.h"
#include "quote_821_local.h"

/* smtp_helo - perform initial handshake with SMTP server */

int     smtp_helo(SMTP_STATE *state)
{
    SMTP_SESSION *session = state->session;
    DELIVER_REQUEST *request = state->request;
    SMTP_RESP *resp;
    int     except;
    char   *lines;
    char   *words;
    char   *word;
    int     n;

    /*
     * Prepare for disaster.
     */
    smtp_timeout_setup(state->session->stream, var_smtp_helo_tmout);
    if ((except = setjmp(smtp_timeout_buf)) != 0)
	return (smtp_stream_except(state, except, "sending HELO"));

    /*
     * Read and parse the server's SMTP greeting banner.
     */
    if (((resp = smtp_chat_resp(state))->code / 100) != 2)
	return (smtp_site_fail(state, resp->code,
			       "host %s refused to talk to me: %s",
			    session->host, translit(resp->str, "\n", " ")));

    /*
     * See if we are talking to ourself. This should not be possible with the
     * way we implement DNS lookups. However, people are known to sometimes
     * screw up the naming service. And, mailer loops are still possible when
     * our own mailer routing tables are mis-configured.
     */
    words = resp->str;
    (void) mystrtok(&words, "- \t\n");
    for (n = 0; (word = mystrtok(&words, " \t\n")) != 0; n++) {
	if (n == 0 && strcasecmp(word, var_myhostname) == 0) {
	    msg_warn("host %s greeted me with my own hostname %s",
		     session->host, var_myhostname);
	    return (smtp_site_fail(state, session->best ? 550 : 450,
				   "mail for %s loops back to myself",
				   request->nexthop));
	} else if (strcasecmp(word, "ESMTP") == 0)
	    state->features |= SMTP_FEATURE_ESMTP;
    }

    /*
     * Return the compliment. Fall back to SMTP if our ESMTP recognition
     * heuristic failed.
     */
    if (state->features & SMTP_FEATURE_ESMTP) {
	smtp_chat_cmd(state, "EHLO %s", var_myhostname);
	if ((resp = smtp_chat_resp(state))->code / 100 != 2)
	    state->features &= ~SMTP_FEATURE_ESMTP;
    }
    if ((state->features & SMTP_FEATURE_ESMTP) == 0) {
	smtp_chat_cmd(state, "HELO %s", var_myhostname);
	if ((resp = smtp_chat_resp(state))->code / 100 != 2)
	    return (smtp_site_fail(state, resp->code,
				   "host %s refused to talk to me: %s",
				   session->host,
				   translit(resp->str, "\n", " ")));
    }

    /*
     * Pick up some useful features offered by the SMTP server. XXX Until we
     * have a portable routine to convert from string to off_t with proper
     * overflow detection, ignore the message size limit advertised by the
     * SMTP server. Otherwise, we might do the wrong thing when the server
     * advertises a really huge message size limit.
     */
    lines = resp->str;
    (void) mystrtok(&lines, "\n");
    while ((words = mystrtok(&lines, "\n")) != 0) {
	if (mystrtok(&words, "- ") && (word = mystrtok(&words, " \t")) != 0) {
	    if (strcasecmp(word, "8BITMIME") == 0)
		state->features |= SMTP_FEATURE_8BITMIME;
	    else if (strcasecmp(word, "PIPELINING") == 0)
		state->features |= SMTP_FEATURE_PIPELINING;
	    else if (strcasecmp(word, "SIZE") == 0)
		state->features |= SMTP_FEATURE_SIZE;
	}
    }
    if (msg_verbose)
	msg_info("server features: 0x%x", state->features);
    return (0);
}

/* smtp_xfer - send a batch of envelope information and the message data */

int     smtp_xfer(SMTP_STATE *state)
{
    DELIVER_REQUEST *request = state->request;
    SMTP_SESSION *session = state->session;
    SMTP_RESP *resp;
    RECIPIENT *rcpt;
    int     curr;
    int     nrcpt;
    int     except;
    int     rec_type;
    int     prev_type = 0;

    /*
     * Send RSET when we already sent a message over this same connection.
     * Otherwise, the "CheckPoint FireWall-1 secure SMTP server" will
     * complain with "503 Sender already specified". Other SMTP servers may
     * have the same bug so we play it safe and send an extra RSET. Ignore
     * the response to RSET, as it may reveal bugs in other SMTP servers. If
     * the server no longer wants to talk to use, we will find out when
     * issuing the next command.
     */
    if (state->done > 0) {
	smtp_timeout_setup(state->session->stream, var_smtp_mail_tmout);
	if ((except = setjmp(smtp_timeout_buf)) != 0)
	    return (smtp_stream_except(state, except, "sending extra RSET"));
	smtp_chat_cmd(state, "RSET");
	(void) smtp_chat_resp(state);
    }

    /*
     * Send the envelope sender address and the optional message size.
     */
#define REWRITE_ADDRESS(addr) { \
	if (*(addr)) { \
	    quote_821_local(state->scratch, addr); \
	    smtp_unalias_addr(state->scratch2, vstring_str(state->scratch)); \
	    myfree(addr); \
	    addr = mystrdup(vstring_str(state->scratch2)); \
	} \
    }

    smtp_timeout_setup(state->session->stream, var_smtp_mail_tmout);
    if ((except = setjmp(smtp_timeout_buf)) != 0)
	return (smtp_stream_except(state, except, "sending MAIL FROM"));

    if (*request->sender)
	REWRITE_ADDRESS(request->sender);
    VSTRING_RESET(state->scratch);
    if (state->features & SMTP_FEATURE_SIZE)
	vstring_sprintf_append(state->scratch, " SIZE=%lu", request->data_size);
    VSTRING_TERMINATE(state->scratch);
    smtp_chat_cmd(state, "MAIL FROM:<%s>%s",
		  request->sender, vstring_str(state->scratch));
    if ((resp = smtp_chat_resp(state))->code / 100 != 2)
	return (smtp_mesg_fail(state, resp->code, "host %s said: %s",
			    session->host, translit(resp->str, "\n", " ")));

    /*
     * Send envelope recipient addresses. Do not send more than the permitted
     * number of recipients at a time. Some SMTP servers violate the protocol
     * and disconnect when their recipient count is exceeded. One example is
     * smap. The workaround is to artificially limit the number of
     * recipients.
     * 
     * For some time this SMTP client interpreted 552 as "too many recipients"
     * (see RFC 821), which suggests the recipient can be tried again later.
     * Unfortunately, someone was using 552 for "this user's mailbox is
     * full", causing a deadly embrace between the Postfix SMTP client and
     * the remote SMTP server.
     * 
     * The lesson here seems to be: in case of problems, draw the most
     * pessimistic conclusion permitted by the standard; retry an operation
     * only when you really have to, or you will be burnt.
     */
    smtp_timeout_setup(state->session->stream, var_smtp_rcpt_tmout);
    if ((except = setjmp(smtp_timeout_buf)) != 0)
	return (smtp_stream_except(state, except, "sending RCPT TO"));

    for (nrcpt = 0, curr = state->done; curr < request->rcpt_list.len; curr++) {
	rcpt = request->rcpt_list.info + curr;
	REWRITE_ADDRESS(rcpt->address);
	smtp_chat_cmd(state, "RCPT TO:<%s>", rcpt->address);
	if ((resp = smtp_chat_resp(state))->code / 100 == 2) {
	    ++nrcpt;
	} else {
	    smtp_rcpt_fail(state, resp->code, rcpt, "host %s said: %s",
			   session->host, translit(resp->str, "\n", " "));
	}
    }

    /*
     * When no deliverable recipient was found, abort the mail transfer
     * dialog. We do not care about the response to RSET because the next
     * command will terminate the session anyway.
     */
    if (nrcpt == 0) {
	smtp_chat_cmd(state, "RSET");
	(void) smtp_chat_resp(state);
	state->done = curr;
	if (state->error_mask == 0)
	    smtp_chat_reset(state);
	return (-1);				/* don't loop */
    }

    /*
     * Request data transmission.
     */
    smtp_timeout_setup(state->session->stream, var_smtp_data0_tmout);
    if ((except = setjmp(smtp_timeout_buf)) != 0)
	return (smtp_stream_except(state, except, "sending DATA command"));

    smtp_chat_cmd(state, "DATA");
    if ((resp = smtp_chat_resp(state))->code / 100 != 3)
	return (smtp_mesg_fail(state, resp->code, "host %s said: %s",
			    session->host, translit(resp->str, "\n", " ")));

    /*
     * Copy message. Escape leading dots. XXX What to do in case of a message
     * file read error? If we can't read it we can't bounce it. To make
     * matters worse, the SMTP protocol has no mechanism to cancel a message
     * once transmission has begun. The best solution is to unceremoniously
     * abort the SMTP session, and to wait for better times.
     */
    smtp_timeout_setup(state->session->stream, var_smtp_data1_tmout);
    if ((except = setjmp(smtp_timeout_buf)) != 0)
	return (smtp_stream_except(state, except, "sending message body"));

    if (vstream_fseek(state->src, request->data_offset, SEEK_SET) < 0)
	msg_fatal("seek queue file: %m");

    while ((rec_type = rec_get(state->src, state->scratch, 0)) > 0) {
	if (rec_type != REC_TYPE_NORM && rec_type != REC_TYPE_CONT)
	    break;
	if (prev_type != REC_TYPE_CONT)
	    if (vstring_str(state->scratch)[0] == '.')
		smtp_fputc('.', session->stream);
	if (rec_type == REC_TYPE_CONT)
	    smtp_fwrite(vstring_str(state->scratch),
			VSTRING_LEN(state->scratch),
			session->stream);
	else
	    smtp_fputs(vstring_str(state->scratch),
		       VSTRING_LEN(state->scratch),
		       session->stream);
	prev_type = rec_type;
    }
    if (prev_type == REC_TYPE_CONT)		/* missing newline at end */
	smtp_fputs("", 0, session->stream);
    if (vstream_ferror(state->src))
	msg_fatal("queue file read error");
    if (rec_type != REC_TYPE_XTRA) {
	shutdown(vstream_fileno(state->session->stream), 0);
	return (mark_corrupt(state->src));
    }

    /*
     * Send "end of message" and process the server response. Warn that a
     * message may be sent more than once when the response to "." is not
     * received.
     */
    smtp_timeout_setup(state->session->stream, var_smtp_data2_tmout);
    if ((except = setjmp(smtp_timeout_buf)) != 0)
	return (smtp_stream_except(state, except,
	      "sending end of data -- message may be sent more than once"));

    smtp_chat_cmd(state, ".");
    if ((resp = smtp_chat_resp(state))->code / 100 != 2)
	return (smtp_mesg_fail(state, resp->code, "host %s said: %s",
			    session->host, translit(resp->str, "\n", " ")));

    /*
     * Cross off the recipients that were actually delivered.
     */
    for (nrcpt = state->done; nrcpt < curr; nrcpt++) {
	rcpt = request->rcpt_list.info + nrcpt;
	if (rcpt->offset) {
	    sent(request->queue_id, rcpt->address,
		 session->host, request->arrival_time, "%s", resp->str);
	    deliver_completed(state->src, rcpt->offset);
	    rcpt->offset = 0;
	}
    }
    state->done = curr;
    if (state->error_mask == 0)
	smtp_chat_reset(state);
    return (0);
}

/* smtp_quit - final SMTP protocol handshake */

void    smtp_quit(SMTP_STATE *state)
{

    /*
     * Send QUIT. Be polite enough to let the SMTP server reply, but ignore
     * the response. We're going to terminate the connection regardless.
     */
    smtp_timeout_setup(state->session->stream, var_smtp_quit_tmout);
    if (setjmp(smtp_timeout_buf) == 0) {
	smtp_chat_cmd(state, "QUIT");
	(void) smtp_chat_resp(state);
    }
}
