/*++
/* NAME
/*	smtpd_ipc 3
/* SUMMARY
/*	smtp server to cleanup service
/* SYNOPSIS
/*	#include "smtpd.h"
/*
/*	int	smtpd_ipc_init(state)
/*	SMTPD_STATE *state;
/* DESCRIPTION
/*	This module implements the smtp server to cleanup service
/*	interface.
/*
/*	smtpd_ipc_init() performs initializations and prepares for
/*	piping a message through the cleanup service.
/* BUGS
/*	Each mail program should be able to write messages to an IPC channel.
/* LICENSE
/* .ad
/* .fi
/*	The Secure Mailer license must be distributed with this software.
/* AUTHOR(S)
/*	Wietse Venema
/*	IBM T.J. Watson Research
/*	P.O. Box 704
/*	Yorktown Heights, NY 10598, USA
/*--*/

/* System library. */

#include <sys_defs.h>

/* Utility library. */

#include <msg.h>
#include <mymalloc.h>

/* Global library. */

#include <mail_proto.h>
#include <cleanup_user.h>

/* Application-specific */

#include "smtpd.h"

/* smtpd_ipc_open - open connection to cleanup service */

static void smtpd_ipc_open(SMTPD_STATE *state)
{

    /*
     * Contact the cleanup service until we succeed. Save a copy of the queue
     * ipc ID. Do not request any special cleanup service features. Bad mail
     * will have to be bounced by the sender.
     */
    state->cleanup =
    mail_connect_wait(MAIL_CLASS_PRIVATE, MAIL_SERVICE_CLEANUP);
    if (mail_scan(state->cleanup, "%s", state->buffer) != 1
	|| mail_print(state->cleanup, "%d", CLEANUP_FLAG_NONE) != 0)
	msg_fatal("unable to connect to cleanup service");
    state->queue_id = mystrdup(vstring_str(state->buffer));
}

/* smtpd_ipc_finish - final handshake with cleanup service */

static void smtpd_ipc_finish(SMTPD_STATE *state)
{

    /*
     * When all is well, receive the cleanup service completion status.
     * Otherwise, just lose it.
     */
    if (state->err == CLEANUP_STAT_OK)
	if (mail_scan(state->cleanup, "%d", &state->err) != 1)
	    state->err = CLEANUP_STAT_WRITE;
}

void    smtpd_ipc_init(SMTPD_STATE *state)
{
    state->open_message = smtpd_ipc_open;
    state->finish_message = smtpd_ipc_finish;
}
